@@ -1,4 +1,5 @@
 from typing import List, TypedDict
+
 from typing_extensions import NotRequired
 from web3.types import Middleware
 
@@ -13,8 +14,8 @@ class NetworkConfig(TypedDict):
     """
 
     name: NetworkName
-    txType: int
-    chainId: int
+    tx_type: int
+    chain_id: int
     middlewares: NotRequired[List[Middleware]]
     rpcs: NotRequired[List[str]]
     coin: str
